"""Provide an MQTT client for providing BoneIO MQTT broker."""
import asyncio
import json
import logging
import uuid
from contextlib import AsyncExitStack
from typing import Any, Callable, Optional, Set, Union

import paho.mqtt.client as mqtt
from asyncio_mqtt import Client as AsyncioClient
from asyncio_mqtt import MqttError
from paho.mqtt.properties import Properties
from paho.mqtt.subscribeoptions import SubscribeOptions

from boneio.const import PAHO
from boneio.helper import UniqueQueue

_LOGGER = logging.getLogger(__name__)


class MQTTClient:
    """Represent an MQTT client."""

    def __init__(
        self,
        host: str,
        port: int = 1883,
        **client_options: Any,
    ) -> None:
        """Set up client."""
        self.host = host
        self.port = port
        if "client_id" not in client_options:
            client_options["client_id"] = mqtt.base62(uuid.uuid4().int, padding=22)
        if "logger" not in client_options:
            client_options["logger"] = logging.getLogger(PAHO)
        client_options["clean_session"] = True
        self.client_options = client_options
        self.asyncio_client: AsyncioClient = None
        self.create_client()
        self.reconnect_interval = 1
        self.publish_queue: UniqueQueue = UniqueQueue()

    def create_client(self) -> None:
        """Create the asyncio client."""
        self.asyncio_client = AsyncioClient(
            self.host,
            self.port,
            **self.client_options,
        )

    async def publish(  # pylint:disable=too-many-arguments
        self,
        topic: str,
        payload: Optional[str] = None,
        qos: int = 0,
        retain: bool = False,
        properties: Optional[Properties] = None,
        timeout: float = 10,
    ) -> None:
        """Publish to topic.

        Can raise asyncio_mqtt.MqttError.
        """
        params: dict = {"qos": qos, "retain": retain, "timeout": timeout}
        if payload:
            params["payload"] = payload
        if properties:
            params["properties"] = properties

        _LOGGER.debug("Sending message topic: %s, payload: %s", topic, payload)
        await self.asyncio_client.publish(topic, **params)

    async def subscribe(  # pylint:disable=too-many-arguments
        self,
        topic: str,
        qos: int = 0,
        options: Optional[SubscribeOptions] = None,
        properties: Optional[Properties] = None,
        timeout: float = 10.0,
    ) -> None:
        """Subscribe to topic.

        Can raise asyncio_mqtt.MqttError.
        """
        params: dict = {"qos": qos, "timeout": timeout}
        if options:
            params["options"] = options
        if properties:
            params["properties"] = properties

        await self.asyncio_client.subscribe(topic, **params)

    async def unsubscribe(
        self, topic: str, properties: Optional[Properties] = None, timeout: float = 10.0
    ) -> None:
        """Unsubscribe from topic.

        Can raise asyncio_mqtt.MqttError.
        """
        params: dict = {"timeout": timeout}
        if properties:
            params["properties"] = properties

        await self.asyncio_client.unsubscribe(topic, **params)

    def send_message(self, topic: str, payload: Union[str, dict]) -> None:
        """Send a message from the manager options."""
        to_publish = (topic, json.dumps(payload) if type(payload) == dict else payload)
        self.publish_queue.put_nowait(to_publish)

    async def _handle_publish(self) -> None:
        """Publish messages as they are put on the queue."""
        while True:
            to_publish: tuple = await self.publish_queue.get()
            await self.publish(*to_publish)
            self.publish_queue.task_done()

    async def start_client(self, manager: any) -> None:
        """Start the client with the manager."""
        # Reconnect automatically until the client is stopped.
        while True:
            try:
                await self._subscribe_manager(manager)
            except MqttError as err:
                self.reconnect_interval = min(self.reconnect_interval * 2, 900)
                _LOGGER.error(
                    "MQTT error: %s. Reconnecting in %s seconds",
                    err,
                    self.reconnect_interval,
                )
                await asyncio.sleep(self.reconnect_interval)
                self.create_client()  # reset connect/reconnect futures

    async def _subscribe_manager(self, manager: any) -> None:
        """Connect and subscribe to manager topics + host stats."""
        async with AsyncExitStack() as stack:
            tasks: Set[asyncio.Task] = set()

            # Connect to the MQTT broker.
            await stack.enter_async_context(self.asyncio_client)
            # Reset the reconnect interval after successful connection.
            self.reconnect_interval = 1

            publish_task = asyncio.create_task(self._handle_publish())
            tasks.add(publish_task)

            # Messages that doesn't match a filter will get logged and handled here.
            messages = await stack.enter_async_context(
                self.asyncio_client.unfiltered_messages()
            )

            messages_task = asyncio.create_task(
                handle_messages(messages, manager.receive_message)
            )
            tasks.add(messages_task)

            topic = f"{manager.relay_topic}"
            await self.subscribe(topic)

            # Wait for everything to complete (or fail due to, e.g., network errors).
            await asyncio.gather(*tasks)


async def handle_messages(messages: Any, callback: Callable[[str, str], None]) -> None:
    """Handle messages with callback."""
    async for message in messages:
        payload = message.payload.decode()
        _LOGGER.debug("Received message topic: %s, payload: %s", message.topic, payload)
        callback(message.topic, payload)



"""Sensor module."""
from boneio.sensor.temp import LM75Sensor, MCP9808Sensor
from boneio.sensor.gpio import GpioInputSensor
from boneio.sensor.adc import initialize_adc, GpioADCSensor

__all__ = [
    "LM75Sensor",
    "MCP9808Sensor",
    "GpioInputSensor",
    "initialize_adc",
    "GpioADCSensor",
]



import asyncio
from boneio.modbus import Modbus
import logging
import time
from typing import Any, Callable, List, Optional, Set, Union

from adafruit_mcp230xx.mcp23017 import MCP23017
from board import SCL, SDA
from busio import I2C

from boneio.const import (
    ACTION,
    ACTIONS,
    ADDRESS,
    BONEIO,
    GPIO,
    HA_TYPE,
    OUTPUT_TYPE,
    HOMEASSISTANT,
    ID,
    INIT_SLEEP,
    INPUT,
    KIND,
    LM75,
    MCP,
    MCP_ID,
    MCP_TEMP_9808,
    OFF,
    ON,
    ONLINE,
    OUTPUT,
    PIN,
    RELAY,
    SDM630,
    STATE,
    ClickTypes,
    InputTypes,
    SENSOR,
    SHOW_HA,
    BINARY_SENSOR,
    INPUT_SENSOR,
    UART,
    UPDATE_INTERVAL,
    UARTS,
)
from boneio.helper import (
    HostData,
    ha_relay_availabilty_message,
    ha_input_availabilty_message,
    ha_adc_sensor_availabilty_message,
    ha_sensor_temp_availabilty_message,
    ha_binary_sensor_availabilty_message,
    host_stats,
    GPIOInputException,
    I2CError,
)
from boneio.input import GpioInputButton

from boneio.relay import GpioRelay, MCPRelay
from boneio.sensor import GpioInputSensor

_LOGGER = logging.getLogger(__name__)


class Manager:
    """Manager to communicate MQTT with GPIO inputs and outputs."""

    def __init__(
        self,
        send_message: Callable[[str, Union[str, dict]], None],
        relay_pins: List = [],
        input_pins: List = [],
        sensors: dict = {},
        topic_prefix: str = BONEIO,
        modbus: dict = None,
        ha_discovery: bool = True,
        ha_discovery_prefix: str = HOMEASSISTANT,
        mcp23017: Optional[List] = None,
        oled: bool = False,
        adc_list: Optional[List] = None,
    ) -> None:
        """Initialize the manager."""
        _LOGGER.info("Initializing manager module.")
        self._loop = asyncio.get_event_loop()
        self._host_data = None
        self._ha_discovery = ha_discovery

        self.send_message = send_message
        self._topic_prefix = topic_prefix
        self.relay_topic = f"{topic_prefix}/{RELAY}/+/set"
        self._input_pins = input_pins
        self._i2cbusio = I2C(SCL, SDA)
        self._mcp = {}
        self._grouped_outputs = {}
        self._oled = None
        self._tasks: List[asyncio.Task] = []
        self._temp_sensors = []
        self._modbus = None
        if modbus and modbus.get(UART) in UARTS:
            self._modbus = Modbus(UARTS[modbus.get(UART)])

        def create_temp_sensor(sensor_type: str, temp_def: dict = {}):
            """Create LM sensor in manager."""
            if sensor_type == LM75:
                from .sensor import LM75Sensor as TempSensor
            elif sensor_type == MCP_TEMP_9808:
                from .sensor import MCP9808Sensor as TempSensor
            else:
                return
            name = temp_def.get(ID)
            id = name.replace(" ", "")
            try:
                temp_sensor = TempSensor(
                    id=id,
                    name=name,
                    i2c=self._i2cbusio,
                    address=temp_def[ADDRESS],
                    send_message=self.send_message,
                    topic_prefix=topic_prefix,
                )
                self.send_ha_autodiscovery(
                    id=id,
                    name=name,
                    ha_type=SENSOR,
                    ha_discovery_prefix=ha_discovery_prefix,
                    availabilty_msg_func=ha_sensor_temp_availabilty_message,
                )
                self._tasks.append(asyncio.create_task(temp_sensor.send_state()))
                self._temp_sensors.append(temp_sensor)
            except I2CError as err:
                _LOGGER.error("Can't configure Temp sensor. %s", err)
                pass

        def create_mcp23017():
            """Create MCP23017."""
            for mcp in mcp23017:
                id = mcp[ID] or mcp[ADDRESS]
                try:
                    self._mcp[id] = MCP23017(i2c=self._i2cbusio, address=mcp[ADDRESS])
                    sleep_time = mcp.get(INIT_SLEEP, 0)
                    _LOGGER.debug(
                        f"Sleeping for {sleep_time}s while MCP {id} is initializing."
                    )
                    time.sleep(sleep_time)
                    self._grouped_outputs[id] = {}
                except TimeoutError as err:
                    _LOGGER.error("Can't connect to MCP %s. %s", id, err)
                    pass

        def create_adc():
            """Create ADC sensor."""
            from boneio.sensor import initialize_adc, GpioADCSensor

            initialize_adc()

            # TODO: find what exception can ADC gpio throw.
            for gpio in adc_list:
                name = gpio.get(ID)
                id = name.replace(" ", "")
                pin = gpio[PIN]
                try:
                    adc = GpioADCSensor(
                        id=id,
                        pin=pin,
                        name=name,
                        send_message=self.send_message,
                        topic_prefix=topic_prefix,
                        update_interval=gpio.get(UPDATE_INTERVAL, 60),
                    )
                    self.send_ha_autodiscovery(
                        id=id,
                        name=name,
                        ha_type=SENSOR,
                        ha_discovery_prefix=ha_discovery_prefix,
                        availabilty_msg_func=ha_adc_sensor_availabilty_message,
                    )
                    self._tasks.append(asyncio.create_task(adc.send_state()))
                except I2CError as err:
                    _LOGGER.error("Can't configure ADC sensor %s. %s", id, err)
                    pass

            return True

        def create_sdm_630(sensors):
            """Create SDM sensor for each device."""
            from boneio.sensor.sdm630 import Sdm630

            for sensor in sensors:
                name = sensor.get(ID)
                id = name.replace(" ", "")
                try:
                    sdm = Sdm630(
                        modbus=self._modbus,
                        address=sensor[ADDRESS],
                        id=id,
                        name=name,
                        send_message=self.send_message,
                        topic_prefix=topic_prefix,
                        ha_discovery=ha_discovery,
                        ha_discovery_prefix=ha_discovery_prefix,
                        update_interval=sensor.get(UPDATE_INTERVAL, 30),
                    )

                    self._tasks.append(asyncio.create_task(sdm.send_state()))
                except I2CError as err:
                    _LOGGER.error("Can't configure SDM630 sensor %s. %s", id, err)
                    pass

            return True

        if sensors.get(LM75):
            create_temp_sensor(sensor_type=LM75, temp_def=sensors.get(LM75))

        if sensors.get(MCP_TEMP_9808):
            create_temp_sensor(
                sensor_type=MCP_TEMP_9808, temp_def=sensors.get(MCP_TEMP_9808)
            )

        if sensors.get(SDM630) and self._modbus:
            create_sdm_630(sensors=sensors.get(SDM630))

        if mcp23017:
            create_mcp23017()

        if adc_list:
            create_adc()

        def configure_relay(gpio: dict) -> Any:
            """Configure kind of relay. Most common MCP."""
            relay_id = gpio[ID].replace(" ", "")
            if gpio[KIND] == MCP:
                mcp_id = gpio.get(MCP_ID, "")
                mcp = self._mcp.get(mcp_id)
                if not mcp:
                    _LOGGER.error("No such MCP configured!")
                    return
                mcp_relay = MCPRelay(
                    pin=int(gpio[PIN]),
                    id=gpio[ID],
                    send_message=self.send_message,
                    topic_prefix=topic_prefix,
                    mcp=mcp,
                    mcp_id=mcp_id,
                    output_type=gpio[OUTPUT_TYPE],
                    callback=lambda: self._host_data_callback(mcp_id),
                )
                self._grouped_outputs[mcp_id][relay_id] = mcp_relay
                return mcp_relay
            elif gpio[KIND] == GPIO:
                if GPIO not in self._grouped_outputs:
                    self._grouped_outputs[GPIO] = {}
                gpio_relay = GpioRelay(
                    pin=gpio[PIN],
                    id=gpio[ID],
                    send_message=self.send_message,
                    topic_prefix=topic_prefix,
                    callback=lambda: self._host_data_callback(GPIO),
                )
                self._grouped_outputs[GPIO][relay_id] = gpio_relay
                return gpio_relay

        self._output = {
            gpio[ID].replace(" ", ""): configure_relay(gpio) for gpio in relay_pins
        }
        for out in self._output.values():
            if out.output_type != "none":
                self.send_ha_autodiscovery(
                    id=out.id,
                    name=out.name,
                    ha_type=out.output_type,
                    ha_discovery_prefix=ha_discovery_prefix,
                    availabilty_msg_func=ha_relay_availabilty_message,
                )
            self._loop.call_soon_threadsafe(
                self._loop.call_later,
                0.5,
                out.send_state,
            )

        def configure_input(gpio):
            try:
                pin = gpio[PIN]
                input_type = gpio.get(KIND)
                if input_type == SENSOR:
                    GpioInputSensor(
                        pin=pin,
                        press_callback=lambda x, i: self.press_callback(
                            x=x,
                            inpin=i,
                            actions=gpio.get(ACTIONS, {}),
                            input_type=INPUT_SENSOR,
                        ),
                        rest_pin=gpio,
                    )
                    availabilty_msg_func = ha_binary_sensor_availabilty_message
                    ha_type = BINARY_SENSOR
                else:
                    GpioInputButton(
                        pin=pin,
                        press_callback=lambda x, i: self.press_callback(
                            x=x,
                            inpin=i,
                            actions=gpio.get(ACTIONS, {}),
                            input_type=INPUT,
                        ),
                        rest_pin=gpio,
                    )
                    availabilty_msg_func = ha_input_availabilty_message
                    ha_type = SENSOR
                if gpio.get(SHOW_HA, True):
                    self.send_ha_autodiscovery(
                        id=pin,
                        name=gpio.get(ID, pin),
                        ha_type=ha_type,
                        ha_discovery_prefix=ha_discovery_prefix,
                        availabilty_msg_func=availabilty_msg_func,
                    )
            except GPIOInputException as err:
                _LOGGER.error("This PIN %s can't be configured. %s", pin, err)
                pass

        _LOGGER.info("Initializing inputs. This will take a while.")
        for gpio in self._input_pins:
            configure_input(gpio=gpio)

        if oled:
            from boneio.oled import Oled

            self._host_data = HostData(
                output=self._grouped_outputs,
                temp_sensor=self._temp_sensors[0] if self._temp_sensors else None,
                callback=self._host_data_callback,
            )
            for f in host_stats.values():
                self._tasks.append(asyncio.create_task(f(self._host_data)))
            _LOGGER.debug("Gathering host data enabled.")
            try:
                self._oled = Oled(
                    host_data=self._host_data, output_groups=list(self._grouped_outputs)
                )
            except (GPIOInputException, I2CError) as err:
                _LOGGER.error("Can't configure OLED display. %s", err)

        self.send_message(topic=f"{topic_prefix}/{STATE}", payload=ONLINE)
        _LOGGER.info("BoneIO manager is ready.")

    def _host_data_callback(self, type: str):
        if self._oled:
            self._oled.handle_data_update(type)

    def get_tasks(self) -> Set[asyncio.Task]:
        return self._tasks

    def press_callback(
        self, x: ClickTypes, inpin: str, actions: dict, input_type: InputTypes = INPUT
    ) -> None:
        """Press callback to use in input gpio.
        If relay input map is provided also toggle action on relay."""
        topic = f"{self._topic_prefix}/{input_type}/{inpin}"
        self.send_message(topic=topic, payload=x)
        action = actions.get(x)
        if action:
            if action[ACTION] == OUTPUT:
                """For now only output type is supported"""
                relay = self._output.get(action[PIN].replace(" ", ""))
                if relay:
                    relay.toggle()
        # This is similar how Z2M is clearing click sensor.
        self._loop.call_soon_threadsafe(self.send_message, topic, "")

    def send_ha_autodiscovery(
        self,
        id: str,
        name: str,
        ha_discovery_prefix: str,
        ha_type: str,
        availabilty_msg_func: Callable,
        topic_prefix: str = None,
        **kwargs,
    ) -> None:
        """Send HA autodiscovery information for each relay."""
        if not self._ha_discovery:
            return
        if not topic_prefix:
            topic_prefix = self._topic_prefix
        msg = availabilty_msg_func(topic=topic_prefix, id=id, name=name, **kwargs)
        topic = f"{ha_discovery_prefix}/{ha_type}/{topic_prefix}/{id}/config"
        _LOGGER.debug("Sending HA discovery for %s, %s.", ha_type, name)
        self.send_message(topic=topic, payload=msg)

    def receive_message(self, topic: str, message: str) -> None:
        """Callback for receiving action from Mqtt."""
        extracted_relay = topic.replace(f"{self._topic_prefix}/{RELAY}/", "").replace(
            "/set", ""
        )
        target_device = self._output.get(extracted_relay)
        if target_device:
            if message == ON:
                target_device.turn_on()
            elif message == OFF:
                target_device.turn_off()

    @property
    def output(self) -> dict:
        """Get list of output."""
        return self._output



import asyncio
import logging
import os
from functools import wraps

import click
from colorlog import ColoredFormatter

from boneio.const import (
    ADC,
    ENABLED,
    HA_DISCOVERY,
    HOST,
    INPUT,
    LM75,
    MCP23017,
    MCP_TEMP_9808,
    MODBUS,
    MQTT,
    OLED,
    OUTPUT,
    PAHO,
    PASSWORD,
    PYMODBUS,
    SDM630,
    TOPIC_PREFIX,
    USERNAME,
)
from boneio.helper import load_config_from_file
from boneio.manager import Manager
from boneio.mqtt_client import MQTTClient
from boneio.version import __version__

_LOGGER = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)
fmt = "%(asctime)s %(levelname)s (%(threadName)s) [%(name)s] %(message)s"
datefmt = "%Y-%m-%d %H:%M:%S"
colorfmt = f"%(log_color)s{fmt}%(reset)s"
logging.getLogger().handlers[0].setFormatter(
    ColoredFormatter(
        colorfmt,
        datefmt=datefmt,
        reset=True,
        log_colors={
            "DEBUG": "cyan",
            "INFO": "green",
            "WARNING": "yellow",
            "ERROR": "red",
            "CRITICAL": "red",
        },
    )
)

MAINPATH = os.path.join(os.path.dirname(__file__))


def add_options(options):
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options


def coro(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        return asyncio.run(f(*args, **kwargs))

    return wrapper


@click.group(no_args_is_help=True)
@click.pass_context
@click.version_option(__version__)
@coro
async def cli(ctx):
    """A tool to run commands."""
    pass


_options = [
    click.option(
        "-d",
        "--debug",
        default=False,
        count=True,
        help="Set Debug mode. Single debug is debug of this lib. Second d is debug of aioxmpp as well.",
    ),
    click.option(
        "--config",
        "-c",
        type=str,
        default="./config.yaml",
        help="Config yaml file. Default to ./config.yaml",
    ),
    click.option(
        "--mqttpassword",
        envvar="MQTTPASS",
        type=str,
        help="Mqtt password. To use as ENV named MQTTPASS",
    ),
]


@cli.command()
@add_options(_options)
@click.pass_context
@coro
async def run(ctx, debug: int, config: str, mqttpassword: str = ""):
    """Run BoneIO."""
    _LOGGER.info("BoneIO starting.")
    if debug == 0:
        logging.getLogger().setLevel(logging.INFO)
    if debug > 0:
        logging.getLogger().setLevel(logging.DEBUG)

        _LOGGER.info("Debug mode active")
        _LOGGER.debug(f"Lib version is {__version__}")
    if debug > 1:
        logging.getLogger(PAHO).setLevel(logging.DEBUG)
        logging.getLogger(PYMODBUS).setLevel(logging.DEBUG)
        logging.getLogger("pymodbus.client").setLevel(logging.DEBUG)
    else:
        logging.getLogger(PAHO).setLevel(logging.WARN)
    _config = load_config_from_file(config_file=config)
    if not _config:
        return
    _LOGGER.info("Connecting to MQTT.")
    client = MQTTClient(
        host=_config[MQTT][HOST],
        username=_config[MQTT].get(USERNAME),
        password=_config[MQTT].get(PASSWORD, mqttpassword),
    )

    manager = Manager(
        send_message=client.send_message,
        topic_prefix=_config[MQTT][TOPIC_PREFIX],
        relay_pins=_config.get(OUTPUT, []),
        input_pins=_config.get(INPUT, []),
        ha_discovery=_config[MQTT][HA_DISCOVERY][ENABLED],
        ha_discovery_prefix=_config[MQTT][HA_DISCOVERY][TOPIC_PREFIX],
        sensors={
            LM75: _config.get(LM75),
            MCP_TEMP_9808: _config.get(MCP_TEMP_9808),
            SDM630: _config.get(SDM630),
        },
        mcp23017=_config.get(MCP23017, []),
        modbus=_config.get(MODBUS),
        oled=_config.get(OLED),
        adc_list=_config.get(ADC, []),
    )
    tasks = set()
    tasks.update(manager.get_tasks())
    tasks.add(client.start_client(manager))
    await asyncio.gather(*tasks)


if __name__ == "__main__":
    asyncio.get_event_loop().run_until_complete(cli())



import logging
from typing import Any
import os
from cerberus import Validator
from boneio.const import ID, OUTPUT
from yaml import load, YAMLError, SafeLoader

schema_file = os.path.join(os.path.dirname(__file__), "../schema.yaml")
_LOGGER = logging.getLogger(__name__)


class Loader(SafeLoader):
    """Loader which support for include in yaml files."""

    def __init__(self, stream):

        self._root = os.path.split(stream.name)[0]

        super(Loader, self).__init__(stream)

    def include(self, node):

        filename = os.path.join(self._root, self.construct_scalar(node))

        with open(filename, "r") as f:
            return load(f, Loader)


Loader.add_constructor("!include", Loader.include)


def load_yaml_file(filename: str) -> Any:
    with open(filename, "r") as stream:
        try:
            return load(stream, Loader=Loader)
        except YAMLError as exception:
            raise exception


def load_config_from_string(config_yaml: str):
    schema = load_yaml_file(schema_file)
    v = CustomValidator(schema, purge_unknown=True)
    return v.normalized(config_yaml)


def load_config_from_file(config_file: str):
    config_yaml = load_yaml_file(config_file)
    if not config_yaml:
        _LOGGER.info("Missing file.")
        return None
    return load_config_from_string(config_yaml)


class CustomValidator(Validator):
    """Custom validator of cerberus"""

    def _check_with_output_id_uniqueness(self, field, value):
        """Check if outputs ids are unique."""
        all_ids = [x[ID] for x in self.document[OUTPUT]]
        if len(all_ids) != len(set(all_ids)):
            self._error(field, "Output IDs are not unique.")

    def _normalize_coerce_to_bool(self, value):
        return True



import logging
import asyncio
from struct import unpack
from collections import namedtuple

from boneio.const import SDM630, SENSOR, STATE, ONLINE, OFFLINE, BASE, LENGTH, REGISTERS
from boneio.helper import BasicMqtt
from boneio.helper.ha_discovery import sdm630_availabilty_message

_LOGGER = logging.getLogger(__name__)


def float32(result, base, addr):
    """Read Float value from register."""
    low = result.getRegister(addr - base)
    high = result.getRegister(addr - base + 1)
    data = bytearray(4)
    data[0] = high & 0xFF
    data[1] = high >> 8
    data[2] = low & 0xFF
    data[3] = low >> 8

    val = unpack("f", bytes(data))

    return val[0]


RegistryEntry = namedtuple(
    "RegistryEntry", "sensor_id address unit_of_measurement state_class device_class"
)

REGISTER_BASE = [
    {
        BASE: 0x00,
        LENGTH: 60,
        REGISTERS: [
            RegistryEntry("Voltage_Phase1", 0x0000, "V", "measurement", "voltage"),
            RegistryEntry("Voltage_Phase2", 0x0002, "V", "measurement", "voltage"),
            RegistryEntry("Voltage_Phase3", 0x0004, "V", "measurement", "voltage"),
            RegistryEntry("Current_Phase1", 0x0006, "A", "measurement", "current"),
            RegistryEntry("Current_Phase2", 0x0008, "A", "measurement", "current"),
            RegistryEntry("Current_Phase3", 0x000A, "A", "measurement", "current"),
            RegistryEntry("Power_Phase1", 0x000C, "W", "measurement", "power"),
            RegistryEntry("Power_Phase2", 0x000E, "W", "measurement", "power"),
            RegistryEntry("Power_Phase3", 0x0010, "W", "measurement", "power"),
            RegistryEntry("Power_VA_Phase1", 0x0012, "VA", "measurement", None),
            RegistryEntry("Power_VA_Phase2", 0x0014, "VA", "measurement", None),
            RegistryEntry("Power_VA_Phase3", 0x0016, "VA", "measurement", None),
            RegistryEntry("Power_VAr_Phase1", 0x0018, "var", "measurement", None),
            RegistryEntry("Power_VAr_Phase2", 0x001A, "var", "measurement", None),
            RegistryEntry("Power_VAr_Phase3", 0x001C, "var", "measurement", None),
            RegistryEntry("Power_Factor_Phase1", 0x001E, "%", "measurement", None),
            RegistryEntry("Power_Factor_Phase2", 0x0020, "%", "measurement", None),
            RegistryEntry("Power_Factor_Phase3", 0x0022, "%", "measurement", None),
            RegistryEntry("Phase_Angle_Phase1", 0x0024, "°", "measurement", None),
            RegistryEntry("Phase_Angle_Phase2", 0x0026, "°", "measurement", None),
            RegistryEntry("Phase_Angle_Phase3", 0x0028, "°", "measurement", None),
            RegistryEntry(
                "Line_to_Neutral_AVG_Volts", 0x002A, "V", "measurement", None
            ),
            RegistryEntry("AVG_Current", 0x002E, "A", "measurement", "current"),
            RegistryEntry("Sum_Current", 0x0030, "A", "measurement", "current"),
            RegistryEntry("Total_Power_Watt", 0x0034, "W", "measurement", "power"),
            RegistryEntry("Total_Power_VA", 0x0038, "VA", "measurement", None),
        ],
    },
    {
        BASE: 0x003C,
        LENGTH: 48,
        REGISTERS: [
            RegistryEntry("System_VAr_Total", 0x003C, "var", "measurement", None),
            RegistryEntry(
                "System_Power_Factor_Total", 0x003E, "%", "measurement", None
            ),
            RegistryEntry("System_PhaseAngle", 0x0042, "°", "measurement", None),
            RegistryEntry("Frequency", 0x0046, "Hz", "measurement", None),
            RegistryEntry(
                "Import_Energy_kWh_Total", 0x0048, "kWh", "total_increasing", "energy"
            ),
            RegistryEntry(
                "Export_Energy_kWh_Total", 0x004A, "kWh", "total_increasing", "energy"
            ),
            RegistryEntry(
                "Import_Energy_kVArh_Total", 0x004C, "kvArh", "total_increasing", None
            ),
            RegistryEntry(
                "Export_Energy_kVArh_Total", 0x004E, "kvArh", "total_increasing", None
            ),
        ],
    },
    {
        BASE: 0x0156,
        LENGTH: 4,
        REGISTERS: [
            RegistryEntry("kWh_Total", 0x0156, "kWh", "total_increasing", "energy"),
            RegistryEntry("kvarh_Total", 0x0158, "kvArh", "total_increasing", None),
        ],
    },
]


class Sdm630(BasicMqtt):
    """Represent Sdm630 sensor in BoneIO."""

    SensorClass = None
    DefaultName = SDM630

    def __init__(
        self,
        modbus,
        address: str,
        ha_discovery_prefix: str,
        topic_prefix: str,
        ha_discovery: bool = False,
        id: str = DefaultName,
        update_interval: int = 60,
        **kwargs,
    ):
        """Initialize Sdm630 class."""
        super().__init__(
            id=id or address, topic_type=SENSOR, topic_prefix=topic_prefix, **kwargs
        )
        self._topic_prefix = topic_prefix
        self._modbus = modbus
        self._address = address
        self._ha_discovery = ha_discovery
        self._discovery_sent = False
        self._ha_discovery_prefix = ha_discovery_prefix
        self._update_interval = update_interval

    def _send_ha_autodiscovery(
        self, id: str, sdm_name: str, sensor_id: str, **kwargs
    ) -> None:
        """Send HA autodiscovery information for each SDM sensor."""
        _LOGGER.debug("Sending HA discovery for sensor %s %s.", sdm_name, sensor_id)
        self._send_message(
            topic=(
                f"{self._ha_discovery_prefix}/{SENSOR}/{self._topic_prefix}{id}"
                f"/{id}{sensor_id.replace('_', '').lower()}/config"
            ),
            payload=sdm630_availabilty_message(
                topic=self._topic_prefix,
                id=id,
                name=sdm_name,
                sensor_id=sensor_id,
                **kwargs,
            ),
        )

    def _send_discovery_for_all_registers(self, register: int = 0) -> bool:
        """Send discovery message to HA for each register."""
        if register > 0:
            for data in REGISTER_BASE:
                for register in data["registers"]:
                    kwargs = {
                        "unit_of_measurement": getattr(register, "unit_of_measurement"),
                        "state_class": getattr(register, "state_class"),
                        "value_template": "{{ value | round(2)}}",
                        "sensor_id": getattr(register, "sensor_id"),
                    }
                    if getattr(register, "device_class"):
                        kwargs["device_class"] = getattr(register, "device_class")
                    self._send_ha_autodiscovery(
                        id=self._id, sdm_name=self._name, **kwargs
                    )
            return True
        return False

    async def send_state(self) -> None:
        """Fetch state periodically and send to MQTT."""
        register = self._modbus.read_single_register(unit=self._address, address=0)
        if not self._discovery_sent and self._ha_discovery:
            self._discovery_sent = self._send_discovery_for_all_registers(register)
        while True:
            payload_online = OFFLINE
            for data in REGISTER_BASE:
                values = self._modbus.read_multiple_registers(
                    unit=self._address, address=data[BASE], count=data[LENGTH]
                )
                if payload_online == OFFLINE:
                    payload_online = ONLINE
                    self._send_message(
                        topic=f"{self._topic_prefix}/{self._id}{STATE}",
                        payload=payload_online,
                    )
                for register in data["registers"]:
                    self._send_message(
                        topic=f'{self._send_topic}/{getattr(register, "sensor_id")}',
                        payload=float32(
                            values, data[BASE], getattr(register, "address")
                        ),
                    )
            await asyncio.sleep(self._update_interval)



from boneio.version import __version__


from boneio.const import INPUT, OFF, ON, RELAY, STATE, SENSOR, INPUT_SENSOR


def ha_relay_availabilty_message(id: str, name: str, topic: str = "boneIO"):
    """Create availability topic for HA."""
    return {
        "availability": [{"topic": f"{topic}/{STATE}"}],
        "command_topic": f"{topic}/relay/{id}/set",
        "device": {
            "identifiers": [topic],
            "manufacturer": "boneIO",
            "model": "boneIO Relay Board",
            "name": f"boneIO {topic}",
            "sw_version": __version__,
        },
        "name": name,
        "payload_off": OFF,
        "payload_on": ON,
        "state_topic": f"{topic}/{RELAY}/{id}",
        "unique_id": f"{topic}{RELAY}{id}",
        "value_template": "{{ value_json.state }}",
    }


def ha_availabilty_message(
    id: str, name: str, topic: str = "boneIO", sensor_type: str = INPUT
):
    """Create availability topic for HA."""
    return {
        "availability": [{"topic": f"{topic}/{STATE}"}],
        "device": {
            "identifiers": [topic],
            "manufacturer": "boneIO",
            "model": "boneIO Relay Board",
            "name": f"boneIO {topic}",
            "sw_version": __version__,
        },
        "icon": "mdi:gesture-double-tap",
        "name": name,
        "state_topic": f"{topic}/{sensor_type}/{id}",
        "unique_id": f"{topic}{sensor_type}{id}",
    }


def ha_input_availabilty_message(**kwargs):
    return ha_availabilty_message(sensor_type=INPUT, **kwargs)


def ha_adc_sensor_availabilty_message(**kwargs):
    msg = ha_availabilty_message(sensor_type=SENSOR, **kwargs)
    msg["unit_of_measurement"] = "V"
    msg["device_class"] = "voltage"
    msg["state_class"] = "measurement"
    return msg


def ha_sensor_availabilty_message(unit_of_measurement: str = None, **kwargs):
    msg = ha_availabilty_message(sensor_type=SENSOR, **kwargs)
    if not unit_of_measurement:
        return msg


def ha_binary_sensor_availabilty_message(id: str, name: str, topic: str = "boneIO"):
    """Create availability topic for HA."""
    return {
        "availability": [{"topic": f"{topic}/{STATE}"}],
        "device": {
            "identifiers": [topic],
            "manufacturer": "boneIO",
            "model": "boneIO Relay Board",
            "name": f"boneIO {topic}",
            "sw_version": __version__,
        },
        "payload_on": "pressed",
        "payload_off": "released",
        "name": name,
        "state_topic": f"{topic}/{INPUT_SENSOR}/{id}",
        "unique_id": f"{topic}{INPUT_SENSOR}{id}",
    }


def ha_sensor_temp_availabilty_message(id: str, name: str, topic: str = "boneIO"):
    """Create availability topic for HA."""
    return {
        "availability": [{"topic": f"{topic}/{STATE}"}],
        "device": {
            "identifiers": [topic],
            "manufacturer": "boneIO",
            "model": "boneIO Relay Board",
            "name": f"boneIO {topic}",
            "sw_version": __version__,
        },
        "name": name,
        "state_topic": f"{topic}/{SENSOR}/{id}",
        "unique_id": f"{topic}{SENSOR}{id}",
        "device_class": "temperature",
        "state_class": "measurement",
        "unit_of_measurement": "°C",
        "value_template": "{{ value_json.state }}",
    }


def sdm630_availabilty_message(
    id: str,
    sensor_id: str,
    name: str,
    topic: str = "sdm630",
    sensor_type: str = SENSOR,
    **kwargs,
):
    """Create SDM630 availability topic for HA."""
    return {
        "availability": [{"topic": f"{topic}/{id}{STATE}"}],
        "device": {
            "identifiers": [id],
            "manufacturer": "boneIO",
            "model": "SDM630",
            "name": f"boneIO {name.upper()}",
            "sw_version": __version__,
        },
        "name": sensor_id,
        "state_topic": f"{topic}/{sensor_type}/{id}/{sensor_id}",
        "unique_id": f"{topic}{sensor_id.replace('_', '').lower()}",
        **kwargs,
    }



"""MCP23017 Relay module."""

import logging
from adafruit_mcp230xx.mcp23017 import MCP23017
from boneio.relay.basic import BasicRelay

_LOGGER = logging.getLogger(__name__)


class MCPRelay(BasicRelay):
    """Represents MCP Relay output"""

    def __init__(self, pin: int, mcp: MCP23017, mcp_id: str, **kwargs) -> None:
        """Initialize MCP relay."""
        super().__init__(**kwargs)
        self._pin_id = pin
        self._pin = mcp.get_pin(pin)
        self._pin.switch_to_output(value=True)
        self._mcp_id = mcp_id
        self._pin.value = False
        _LOGGER.debug("Setup MCP with pin %s", self._pin_id)

    @property
    def is_mcp_type(self) -> bool:
        """Check if relay is mcp type."""
        return True

    @property
    def pin_id(self) -> str:
        """Return PIN id."""
        return self._pin_id

    @property
    def mcp_id(self) -> str:
        """Retrieve parent MCP ID."""
        return self._mcp_id

    @property
    def is_active(self) -> bool:
        """Is relay active."""
        return self.pin.value

    @property
    def pin(self) -> str:
        """PIN of the relay"""
        return self._pin

    def turn_on(self) -> None:
        """Call turn on action."""
        self.pin.value = True
        self._loop.call_soon_threadsafe(self.send_state)

    def turn_off(self) -> None:
        """Call turn off action."""
        self.pin.value = False
        self._loop.call_soon_threadsafe(self.send_state)



"""Manage BoneIO onboard temp sensors."""

import asyncio
from adafruit_pct2075 import PCT2075
from adafruit_mcp9808 import MCP9808
from boneio.const import STATE, SENSOR, LM75, MCP_TEMP_9808, TEMPERATURE

from boneio.helper.exceptions import I2CError
from boneio.helper import BasicMqtt


class TempSensor(BasicMqtt):
    """Represent Temp sensor in BoneIO."""

    SensorClass = None
    DefaultName = TEMPERATURE

    def __init__(self, i2c, address, id: str = DefaultName, **kwargs):
        """Initialize Temp class."""
        super().__init__(id=id, topic_type=SENSOR, **kwargs)
        try:
            self._pct = self.SensorClass(i2c_bus=i2c, address=address)
        except ValueError as err:
            raise I2CError(err)

    @property
    def state(self):
        """Give rounded value of temperature."""
        return round(self._pct.temperature, 2)

    async def send_state(self):
        """Fetch temperature periodically and send to MQTT."""
        while True:
            self._send_message(
                topic=self._send_topic,
                payload={STATE: self.state},
            )
            await asyncio.sleep(60)


class LM75Sensor(TempSensor):
    """Represent LM75 sensor in BoneIO."""

    SensorClass = PCT2075
    DefaultName = LM75


class MCP9808Sensor(TempSensor):
    """Represent MCP9808 sensor in BoneIO."""

    SensorClass = MCP9808
    DefaultName = MCP_TEMP_9808
